# copyright 2000, greg landrum
"""

Here we're implementing the Decision Tree stuff found in Chapter 3 of
Tom Mitchell's Machine Learning Book.

"""
from warnings import warn

warn('This module is deprecated and will be removed in the 2024.03 release', DeprecationWarning,
     stacklevel=2)
