//
//  Copyright (C) 2002-2006 Greg Landrum and Rational Discovery LLC
//
//   @@ All Rights Reserved @@
//  This file is part of the RDKit.
//  The contents are covered by the terms of the BSD license
//  which is included in the file license.txt, found at the root
//  of the RDKit source tree.
//
#include <RDGeneral/export.h>
#ifndef _RD_PRIMES_H
#define _RD_PRIMES_H

namespace RDKit {
const int NUM_PRIMES_AVAIL = 1000;
extern int firstThousandPrimes[NUM_PRIMES_AVAIL];
}  // namespace RDKit
#endif
