# copyright 2000, greg landrum
"""

Here is the implementation for K-nearest neighbors

"""
from warnings import warn

warn('This module is deprecated and will be removed in the 2024.03 release', DeprecationWarning,
     stacklevel=2)
