# copyright 2003, Rational Discovery LLC
"""

An implementation of the Naive Bayes Classifier

"""
from warnings import warn

warn('This module is deprecated and will be removed in the 2024.03 release', DeprecationWarning,
     stacklevel=2)

from rdkit import rdBase
